"""
=======
Pharmpy
=======

Pharmpy is a python package for pharmacometrics modeling.

Definitions
===========
"""

__version__ = '0.81.1'

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())
